#include "raycaster.hpp"
#include <iostream>
#include <chrono>
#include <numeric>
#include <functional>
#include "constants.hpp"
#include "stats.hpp"
#include "levelmanager.hpp"
#include "components.hpp"

using namespace components;

void draw_gui(sf::RenderWindow &window, sf::Text &text, Stats &stats) {
  sf::RectangleShape rect({SCREEN_WIDTH - RAY_VIEW_WIDTH, SCREEN_HEIGHT});

  rect.setPosition({0,0});
  rect.setFillColor({50, 50, 50});
  window.draw(rect);

  text.setString(
    fmt::format("FPS\nmean:{:>8.5}\nsdev: {:>8.5}\nmin: {:>8.5}\nmax: {:>8.5}\ncount:{:<10}\n\nVSync? {}\nFR Limit: {}\nDebug? {}\n\nHit R to reset.",
      stats.mean(), stats.stddev(), stats.min, stats.max, stats.n, VSYNC, FRAME_LIMIT, DEBUG_BUILD));
  window.draw(text);
}

Matrix generate_map(TexturePack &textures, GameLevel &level, Point &player_out) {
  auto &tiles = level.map->tiles();
  auto &player = level.world->get_the<Player>();
  auto &player_position = level.world->get<Position>(player.entity);
  player_out = player_position.location;

  return textures.convert_char_to_texture(tiles.$tile_ids);
}

void draw_weapon(sf::RenderWindow &window, sf::Sprite &weapon, float rotation) {
  weapon.setPosition({SCREEN_WIDTH/2,SCREEN_HEIGHT/2});
  weapon.setRotation(sf::degrees(rotation));
  window.draw(weapon);
}

int main() {
  sf::RenderWindow window(sf::VideoMode({SCREEN_WIDTH, SCREEN_HEIGHT}), "Zed's Ray Caster Game Thing");

  sf::Font font{"./assets/text.otf"};
  sf::Text text{font};
  text.setFillColor({255,255,255});
  text.setPosition({10,10});

  Point player{0, 0};
  LevelManager levels;
  GameLevel &cur_level = levels.current();

  TexturePack textures;
  textures.load_tiles();
  textures.load_sprites();

  auto map = generate_map(textures, cur_level, player);

  Point evil_eye_pos{player.x+1, player.y+1};

  Raycaster rayview(window, textures, map, RAY_VIEW_WIDTH, RAY_VIEW_HEIGHT);
  rayview.set_position(RAY_VIEW_X, RAY_VIEW_Y);
  rayview.position_camera(player.x, player.y);
  rayview.init_shaders();
  DinkyECS::Entity evil_ent = rayview.position_sprite(evil_eye_pos, "evil_eye");

  double moveSpeed = 0.1;
  double rotSpeed = 0.1;

  const auto onClose = [&window](const sf::Event::Closed&)
  {
      window.close();
  };

  float rotation = -30.0f;
  Stats stats;

  window.setVerticalSyncEnabled(VSYNC);
  window.setFramerateLimit(FRAME_LIMIT);

  double new_x = evil_eye_pos.x+0.1;
  double new_y = evil_eye_pos.y+0.1;

  while(window.isOpen()) {
    auto start = std::chrono::high_resolution_clock::now();
    rayview.render();
    auto end = std::chrono::high_resolution_clock::now();
    auto elapsed = std::chrono::duration<double>(end - start);
    stats.sample(1/elapsed.count());

    auto weapon_sprite_ptr = rayview.$textures.sword.sprite;
    draw_gui(window, text, stats);
    draw_weapon(window, *weapon_sprite_ptr, rotation);
    window.display();

    if(sf::Keyboard::isKeyPressed(sf::Keyboard::Key::W)) {
      rayview.run(moveSpeed, 1);
    } else if(sf::Keyboard::isKeyPressed(sf::Keyboard::Key::S)) {
      rayview.run(moveSpeed, -1);
    }

    if(sf::Keyboard::isKeyPressed(sf::Keyboard::Key::D)) {
      rayview.rotate(rotSpeed, -1);
    } else if(sf::Keyboard::isKeyPressed(sf::Keyboard::Key::A)) {
      rayview.rotate(rotSpeed, 1);
    }

    if(sf::Keyboard::isKeyPressed(sf::Keyboard::Key::R)) {
      stats.reset();
    }

    if(sf::Keyboard::isKeyPressed(sf::Keyboard::Key::P)) {
      if(rayview.$active_shader == nullptr) {
        rayview.$active_shader = &rayview.$paused;
      } else {
        rayview.$active_shader = nullptr;
      }
    }

    if(sf::Mouse::isButtonPressed(sf::Mouse::Button::Left)) {
      new_x += 0.1;
      new_y += 0.1;
      rayview.$collision.move(evil_eye_pos, {size_t(new_x), size_t(new_y)}, evil_ent);
      evil_eye_pos = {size_t(new_x), size_t(new_y)};
      rayview.$sprites[evil_ent].x = new_x;
      rayview.$sprites[evil_ent].y = new_y;

      rayview.$anim.play(false);
      rotation = -30.0f;
    } else {
      rotation = -10.0f;
    }

    window.handleEvents(onClose);
  }

  return 0;
}
